# os.path is used for test
modules = ['redis', 'huggingface_hub', 'jieba', 'modelscope', 'pandas', 'jwt', 'rank_bm25', 'redisvl', 'datasets',
           'deepspeed', 'fire', 'numpy', 'peft', 'torch', 'transformers', 'faiss', 'flash_attn', 'google', 'lightllm',
           'vllm', 'ChatTTS', 'wandb', 'funasr', 'sklearn', 'torchvision', 'scipy', 'pymilvus', 'sentence_transformers',
           'gradio', 'chromadb', 'nltk', ['PIL', 'Image'], 'httpx', 'bm25s', 'kubernetes', 'pymongo',
           'rapidfuzz', 'FlagEmbedding', 'mcp', 'diffusers', 'pypdf', 'pptx', 'html2text', 'ebooklib', 'docx2txt',
           'zlib', 'struct', 'olefile', 'spacy', 'tarfile', 'boto3', 'botocore', 'paddleocr', 'volcenginesdkarkruntime',
           'zhipuai', 'dashscope', ['mineru', 'cli.common'], 'opensearchpy', ['os', 'path'], 'pkg_resources', 'fastapi',
           ['fsspec', 'implementations.local'], 'bs4', 'uvicorn', 'elasticsearch', 'xml', 'deepdiff', 'mem0', 'memu',
           ['graphrag', 'api', 'config.load_config', 'config.enums', 'cli.index', 'cli.initialize'], 'pyobvector',
           'transformers']
